/* === This file is part of Calamares - <https://github.com/calamares> ===
 *
 *   Copyright 2014, Aurélien Gâteau <agateau@kde.org>
 *   Copyright 2015, Teo Mrnjavac <teo@kde.org>
 *   Copyright 2017, Adriaan de Groot <groot@kde.org>
 *
 *   Calamares is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Calamares is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Calamares. If not, see <http://www.gnu.org/licenses/>.
 */

#include "jobs/CreatePartitionTableJob.h"

#include "JobQueue.h"
#include "GlobalStorage.h"

#include "core/PartitionIterator.h"
#include "core/PartitionInfo.h"

#include "utils/Logger.h"

// KPMcore
#include <core/device.h>
#include <core/partition.h>
#include <core/partitiontable.h>
#include <fs/filesystem.h>
#include <fs/luks.h>
#include <ops/createpartitiontableoperation.h>
#include <util/report.h>

// Qt
#include <QProcess>
#include <QJsonArray>
#include <QJsonObject>

const QString MarinerConfGS = "MARINERCONF";

const QString KeyID = "ID";
const QString KeyName = "Name";
const QString KeyType = "Type";
const QString KeyValue = "Value";
const QString KeyFlags = "Flags";
const QString KeyStart = "Start";
const QString KeyEnd = "End";
const QString KeyFsType = "FsType";
const QString KeyRemoveDocs = "RemoveDocs";
const QString KeyMountOptions = "MountOptions";
const QString KeyMountPoint = "MountPoint";
const QString KeyArtifacts = "Artifacts";
const QString KeyPartitionTableType = "PartitionTableType";
const QString KeyTargetDisk = "TargetDisk";
const QString KeyRawBinaries = "RawBinaries";
const QString KeyPartitions = "Partitions";
const QString KeyPartitionSettings = "PartitionSettings";
const QString KeyRootEncryption = "Encryption";
const QString KeyEnableEncryption = "Enable";
const QString KeyPasswordEncryption = "Password";        
const QString KeySystemConfigs = "SystemConfigs";
const QString KeyMaxSize = "MaxSize";
const QString KeyDisks = "Disks";

// Only support ext4 on LUKS for now
const QString DefaultLuksFS = "ext4";


CreatePartitionTableJob::CreatePartitionTableJob( Device* device, PartitionTable::TableType type )
    : m_device( device )
    , m_type( type )
{
}

QString
CreatePartitionTableJob::prettyName() const
{
    return tr( "Create new %1 partition table on %2." )
            .arg( PartitionTable::tableTypeToName( m_type ) )
            .arg( m_device->deviceNode() );
}

QString CreatePartitionTableJob::prettyDescription() const
{
    return tr( "Create new <strong>%1</strong> partition table on <strong>%2</strong> (%3)." )
            .arg( PartitionTable::tableTypeToName( m_type ).toUpper() )
            .arg( m_device->deviceNode() )
            .arg( m_device->name() );
}


QString
CreatePartitionTableJob::prettyStatusMessage() const
{
    return tr( "Creating new %1 partition table on %2." )
            .arg( PartitionTable::tableTypeToName( m_type ).toUpper() )
            .arg( m_device->deviceNode() );
}


static inline QDebug&
operator <<( QDebug&& s, PartitionIterator& it )
{
    s << ( ( *it ) ? ( *it )->deviceNode() : QString( "<null device>" ) );
    return s;
}


// labelize creates a label for partition X
// for use in the mariner configuration JSON
// The label is guaranteed to be unique based on path;
// Partitions with empty path obtain a unique number
// EACH TIME the function is called - caller should
// save the result and re-use when referring to partition's label.
QString labelize(Partition* x)
{
    static unsigned uniqueCounter = 0;
    QString mountPoint = PartitionInfo::mountPoint(x);
    if(mountPoint == "/")
    {
        return "rootfs";
    }
    else
    {
        // Report as path without slashes
        QString result = mountPoint.replace("/", "");
        if(result == "")
        {
            // Empty path? Return next unique number.
            return QString::number(uniqueCounter++);
        }
        else
        {
            return result;
        }
    }
}


Calamares::JobResult
CreatePartitionTableJob::exec()
{
    Calamares::GlobalStorage* gs = Calamares::JobQueue::instance()->globalStorage();

    // Obtain preloaded configuration from the store
    QJsonObject marinerConfiguration = gs->value(MarinerConfGS).toJsonObject();
    QJsonArray partitions, partitionSettings;
    QJsonObject disk;
    QJsonObject rootEncryption;

    // Collect partition and basic disk data
    emitPartitionJSONs(partitions, partitionSettings, disk, rootEncryption);

    // Add partition data to the disk configuration
    disk[KeyPartitions] = partitions;
    marinerConfiguration[KeyDisks] = QJsonArray() << disk;

    // Add partition mount information to the configuration
    // There is only one configuration needed
    // and it should be already filled with other information
    QJsonObject systemConf = marinerConfiguration[KeySystemConfigs].toArray()[0].toObject();
    systemConf[KeyPartitionSettings] = partitionSettings;
    systemConf[KeyRootEncryption] = rootEncryption;
    marinerConfiguration[KeySystemConfigs] = QJsonArray() << systemConf;

    gs->remove(MarinerConfGS);
    gs->insert(MarinerConfGS, marinerConfiguration);
    return Calamares::JobResult::ok();
}

//TODO: Multiple disks unsupported!
void CreatePartitionTableJob::emitPartitionJSONs(QJsonArray& partitions,
                                                 QJsonArray& partitionSettings,
                                                 QJsonObject& disk,
                                                 QJsonObject& rootEncryption)
{
    const unsigned MiBSize = 1048576; // in bytes
    PartitionTable* table = m_device->partitionTable();

    disk[KeyPartitionTableType] = table->type() == PartitionTable::TableType::msdos ? "mbr" : "gpt";
    disk[KeyMaxSize] = 0; // Don't bother with reporting real size
    disk[KeyArtifacts] = QJsonValue();
    disk[KeyRawBinaries] = QJsonValue();
    bool targetDiskInitialized = false;

    for (auto it = PartitionIterator::begin(table); it != PartitionIterator::end(table); ++it)
    {
        Partition* x = *it;
        // Unknown = empty space, not a partition - skip.
        if(x->fileSystem().type() == FileSystem::Type::Unknown)
        {
            continue;
        }

        if(!targetDiskInitialized)
        {
            QJsonObject targetDisk;
            targetDisk[KeyType] = "path";
            targetDisk[KeyValue] = x->devicePath();

            disk[KeyTargetDisk] = targetDisk;
            targetDiskInitialized = true;
        }

        QJsonObject currentPartition;
        currentPartition[KeyID] = labelize(x);
        currentPartition[KeyName] = currentPartition[KeyID];
        currentPartition[KeyFlags] = QJsonArray::fromStringList(PartitionTable::flagNames(PartitionInfo::flags(x)));
        currentPartition[KeyStart] = (x->firstByte() / MiBSize);

        // check if we're almost at capacity and set end to 0 in json (full size)
        qint64 threshold = m_device->capacity() / 100; // approx 1% of capacity
        if (x->lastByte() >= m_device->capacity() - threshold)
        {
            currentPartition[KeyEnd] = 0;
        }
        else
        {
            // Since Mariner installer does not use the last MiB,
            // we have to report additional 1MiB per partition end
            currentPartition[KeyEnd] = x->lastByte() / MiBSize + 1;
        }
        currentPartition[KeyFsType] = x->fileSystem().name();

        // Check for root encryption
        if ( x->roles().has( PartitionRole::Luks ) && (labelize(x) == "rootfs"))
        {
            rootEncryption[KeyEnableEncryption] = true;

            const FileSystem& fsRef = x->fileSystem();
            const FS::luks* luksFs = dynamic_cast< const FS::luks* >( &fsRef );
            rootEncryption[KeyPasswordEncryption] = luksFs->passphrase();

            currentPartition[KeyFsType] = DefaultLuksFS;
            
            // This is a device mapper root, append the dmroot flag so we can track it during installation.
            QJsonArray partFlagArray = currentPartition[KeyFlags].toArray();
            partFlagArray.append("dmroot");
            currentPartition[KeyFlags] = partFlagArray;
        }

        partitions << currentPartition;

        // Emit Partition Setting - specifies partition entry
        QJsonObject currentPartitionSetting;
        currentPartitionSetting[KeyRemoveDocs] = false;
        currentPartitionSetting[KeyID] = currentPartition[KeyID];
        currentPartitionSetting[KeyMountOptions] = "";
        currentPartitionSetting[KeyMountPoint] = PartitionInfo::mountPoint(x);
        currentPartitionSetting[KeyArtifacts] = QJsonArray();

        partitionSettings << currentPartitionSetting;
    }
}

void
CreatePartitionTableJob::updatePreview()
{
    // Device takes ownership of its table, but does not destroy the current
    // one when setPartitionTable() is called, so do it ourself
    delete m_device->partitionTable();
    m_device->setPartitionTable( createTable() );
    m_device->partitionTable()->updateUnallocated( *m_device );
}

PartitionTable*
CreatePartitionTableJob::createTable()
{
    cDebug() << "CreatePartitionTableJob::createTable trying to make table for device"
             << m_device->deviceNode();
    return new PartitionTable( m_type,
                               PartitionTable::defaultFirstUsable( *m_device, m_type ),
                               PartitionTable::defaultLastUsable( *m_device, m_type )
                             );
}
