/* === This file is part of Calamares - <https://github.com/calamares> ===
 *
 *   Copyright 2014-2017, Teo Mrnjavac <teo@kde.org>
 *   Copyright 2017, Adriaan de Groot <groot@kde.org>
 *
 *   Calamares is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Calamares is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Calamares. If not, see <http://www.gnu.org/licenses/>.
 */

#include <MarinerInstallerJob.h>

#include "JobQueue.h"
#include "GlobalStorage.h"
#include "utils/Logger.h"
#include "utils/CalamaresUtilsSystem.h"
#include "ExecutionViewStep.h"

#include <QDir>
#include <QThread>
#include <QProcess>

#include <random>
#include <sstream>
#include <time.h>

#ifndef NO_CRYPT_H
#include <crypt.h>
#endif
#include <unistd.h>


MarinerinstallerJob::MarinerinstallerJob( const std::string& ksPath)
    : Calamares::Job()
    , m_ksPath( ksPath )
    , m_currentPercent( 0 )
    , m_currentAction( "Starting Installation")
{
}


QString
MarinerinstallerJob::prettyName() const
{
    return tr("Run installer with kickstart %1" ).arg(QString::fromStdString(m_ksPath));
}


QString
MarinerinstallerJob::prettyStatusMessage() const
{
    return m_currentAction;
}


Calamares::JobResult
MarinerinstallerJob::exec()
{
    Calamares::GlobalStorage* gs = Calamares::JobQueue::instance()->globalStorage();

    const QString Program = "/etc/calamares/mariner-install.sh";
    QProcess installerProc;

    // Forward the installer's stderr to Calamares Stderr so it will be captured
    // in the log file.
    installerProc.setProcessChannelMode(QProcess::ForwardedErrorChannel);

    time_t startTime = time(NULL);
    installerProc.start(Program);
    if (!installerProc.waitForStarted()) {
        return Calamares::JobResult::error(tr("Failed to start %1").arg(Program));
    }

    bool installerFinished = false;
    while(m_currentPercent != 100) {
        // Block until data (stdout) is available to be read.
        //
        // Note, if this returns false then the installer process has already
        // exited before all of the stdout could be processed.
        //
        // To prevent the UI from jumping to the finish screen while the progress bar
        // is not filled in completely, sleep for a short duration and keep consuming the
        // remaining stdout to provide a smooth transition.
        if (!installerProc.waitForReadyRead(-1)) {
            Q_ASSERT(installerProc.waitForFinished(0));
            installerFinished = true;
            QThread::msleep(5);
        }

        // Wait for a whole line or EOF to be available
        char buf[1024];
        qint64 lineLength = installerProc.readLine(buf, sizeof(buf));
        if (lineLength == -1) {
            if (installerFinished) {
                // If the installer has already finished and stdout is still being proceded,
                // there should always be more data until percent is reported to be 100%.
                //
                // If not then there was an error during installation and the loop should be broken.
                break;
            } else {
                // A line is still being written to stdout, and \n hasn't been written yet.
                // Repeat the loop until it is ready.
                continue;
            }
        }

        const QString ProgressKey = "progress:";
        const QString ActionKey = "action:";    
        QString line = QString(buf);
        // Remove the trailing new line character
        line.replace("\n", "");

        if (line.startsWith(ProgressKey)) {
            m_currentPercent = line.remove(0, ProgressKey.size()).toInt();
        } else if (line.startsWith(ActionKey)) {
            m_currentAction = line.remove(0, ActionKey.size());
        }

        // emit progress will invoke an update on both the progress bar and the label
        emit progress( qreal( m_currentPercent / 100.0 ) );
    }

    // Wait for the installer process to exit
    const int WaitIndefinitely = -1;
    installerProc.waitForFinished(WaitIndefinitely);

    time_t endTime = time(NULL);

    if (installerProc.exitCode() == 0) {
        gs->insert( "installDuration", difftime(endTime, startTime) );
        return Calamares::JobResult::ok();
    } else {
        return Calamares::JobResult::error(tr("Internal Error. See /var/log/imager.log for details."));
    }
}
