/* === This file is part of Calamares - <https://github.com/calamares> ===
 *
 *   Copyright 2014-2017, Teo Mrnjavac <teo@kde.org>
 *   Copyright 2017-2018, Adriaan de Groot <groot@kde.org>
 *   Copyright 2019, Collabora Ltd <arnaud.ferraris@collabora.com>
 *
 *   Portions from the Manjaro Installation Framework
 *   by Roland Singer <roland@manjaro.org>
 *   Copyright (C) 2007 Free Software Foundation, Inc.
 *
 *   Calamares is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Calamares is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Calamares. If not, see <http://www.gnu.org/licenses/>.
 */

#include "UsersPage.h"

#include "ui_page_usersetup.h"

#include "MarinerConfJob.h"
#include "MarinerInstallerJob.h"

#include "GlobalStorage.h"
#include "JobQueue.h"
#include "Settings.h"

#include "utils/CalamaresUtilsGui.h"
#include "utils/Logger.h"
#include "utils/Retranslator.h"
#include "utils/String.h"

#include <random>
#include <string>
#include <algorithm>

#include <QBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QRegExp>
#include <QRegExpValidator>
#include <QString>

/** Add an error message and pixmap to a label. */
static inline void
labelError( QLabel* pix, QLabel* label, const QString& message )
{
    label->setText( message );
    pix->setPixmap( CalamaresUtils::defaultPixmap( CalamaresUtils::No, CalamaresUtils::Original, label->size() ) );
}

/** Clear error, indicate OK on a label. */
static inline void
labelOk( QLabel* pix, QLabel* label )
{
    label->clear();
    pix->setPixmap( CalamaresUtils::defaultPixmap( CalamaresUtils::Yes, CalamaresUtils::Original, label->size() ) );
}

QString randomHostname(QString prefix)
{
    const unsigned postfixLength = 12;
    static constexpr auto chars =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";

    std::default_random_engine random(time(0));
    std::uniform_int_distribution<int> dist(0, strlen(chars) - 1);
    std::string result(postfixLength, '\0');
    std::generate_n(begin(result), postfixLength, [&]() { return chars[dist(random)]; });

    return prefix + "-" + QString::fromStdString(result);
}

UsersPage::UsersPage( QWidget* parent )
    : QWidget( parent )
    , ui( new Ui::Page_UserSetup )
    , m_readyFullName( false )
    , m_readyUserName( false )
    , m_readyHostname( false )
    , m_readyPassword( false )
    , m_readyRootPassword( false )
{
    ui->setupUi( this );

    // Connect signals and slots
    connect( ui->textBoxHostname, &QLineEdit::textEdited,
             this, &UsersPage::onHostnameTextEdited );
    connect( ui->textBoxUserName, &QLineEdit::textEdited,
             this, &UsersPage::onUserNameTextEdited );
    connect( ui->textBoxUserPassword, &QLineEdit::textChanged,
             this, &UsersPage::onUserPasswordTextChanged );
    connect( ui->textBoxUserVerifiedPassword, &QLineEdit::textChanged,
             this, &UsersPage::onUserPasswordTextChanged );

    fillSuggestions();
}


UsersPage::~UsersPage()
{
    delete ui;
}


bool
UsersPage::isReady()
{
    return m_readyHostname && m_readyUserName && m_readyPassword;
}


QList< Calamares::job_ptr >
UsersPage::createJobs( const QStringList& defaultGroupsList )
{
    const std::string confPath = "/etc/calamares/unattended_config.json";
    QList< Calamares::job_ptr > jobs;

    if ( !isReady() )
        return jobs;

    Calamares::GlobalStorage* gs = Calamares::JobQueue::instance()->globalStorage();

    Calamares::Job* j = new MarinerconfJob( ui->textBoxHostname->text(),
                                            ui->textBoxUserName->text(),
                                            ui->textBoxUserPassword->text(),
                                            confPath);
    jobs.append(Calamares::job_ptr( j ));

    j = new MarinerinstallerJob(confPath);
    jobs.append(Calamares::job_ptr( j ));

    return jobs;
}


void
UsersPage::onActivate()
{
    ui->textBoxUserName->setFocus();
}

void
UsersPage::fillSuggestions()
{
    const QString DefaultHostnamePrefix = "cbl-mariner";
    ui->textBoxHostname->setText(randomHostname(DefaultHostnamePrefix));
    validateHostnameText(ui->textBoxHostname->text());
}

void
UsersPage::onHostnameTextEdited( const QString& textRef )
{
    validateHostnameText( textRef );
}

void
UsersPage::onUserNameTextEdited( const QString& textRef )
{
    validateUserNameText( textRef );
}

void
UsersPage::validateHostnameText( const QString& textRef )
{
    QString text = textRef;
    QRegExp rx( HOSTNAME_RX );
    QRegExpValidator val( rx );
    int pos = -1;

    if ( text.isEmpty() )
    {
        ui->labelHostnameError->clear();
        ui->labelHostname->clear();
        m_readyHostname= false;
    }
    else if ( text.length() < HOSTNAME_MIN_LENGTH )
    {
        labelError( ui->labelHostname, ui->labelHostnameError,
                    tr( "Your hostname is too short. The minimum length is %1 characters." ).arg(HOSTNAME_MIN_LENGTH));
        m_readyHostname = false;
    }
    else if ( text.length() > HOSTNAME_MAX_LENGTH )
    {
        labelError( ui->labelHostname, ui->labelHostnameError,
                    tr( "Your hostname is too long. The maximum length is %1 characters." ).arg(HOSTNAME_MAX_LENGTH) );
        m_readyHostname = false;
    }
    else if ( val.validate( text, pos ) == QValidator::Invalid )
    {
        labelError( ui->labelHostname, ui->labelHostnameError,
                    tr( "Your hostname contains invalid characters. Only letters, numbers and dashes are allowed." ) );
        m_readyHostname = false;
    }
    else
    {
        labelOk( ui->labelHostname, ui->labelHostnameError );
        m_readyHostname = true;
    }

    emit checkReady( isReady() );
}

void
UsersPage::validateUserNameText( const QString& textRef )
{
    QString text = textRef;
    QRegExp rx( USERNAME_RX );
    QRegExpValidator val( rx );
    int pos = -1;

    // Assume it's invalid, then check
    m_readyUserName= false;

    if ( text.isEmpty() )
    {
        ui->labelUserNameError->clear();
        ui->labelUserName->clear();
    }
    else if ( text.length() < USERNAME_MIN_LENGTH )
    {
        labelError( ui->labelUserName, ui->labelUserNameError,
                    tr( "Your User Name is too short. The minimum length is %1 characters." ).arg(USERNAME_MIN_LENGTH) );
    }
    else if ( text.length() > USERNAME_MAX_LENGTH )
    {
        labelError( ui->labelUserName, ui->labelUserNameError,
                    tr( "Your User Name is too long. The maximum length is %1 characters." ).arg(USERNAME_MAX_LENGTH) );
    }
    else if ( val.validate( text, pos ) == QValidator::Invalid )
    {
        labelError( ui->labelUserName, ui->labelUserNameError,
                    tr( "Your User Name contains invalid characters. Only lower-case letters, numbers and dashes are allowed. Name cannot start with a dash." ) );
    }
    else
    {
        // Username is valid, show okay sign.
        labelOk( ui->labelUserName, ui->labelUserNameError );
        m_readyUserName = true;
    }

    emit checkReady( isReady() );
}


void
UsersPage::onUserPasswordTextChanged( const QString& )
{
    QString pw1 = ui->textBoxUserPassword->text();
    QString pw2 = ui->textBoxUserVerifiedPassword->text();

    // TODO: 3.3: remove empty-check and leave it to passwordRequirements
    if ( pw1.isEmpty() && pw2.isEmpty() )
    {
        ui->labelUserPasswordError->clear();
        ui->labelUserPassword->clear();
        m_readyPassword = false;
    }
    else if ( pw1 != pw2 )
    {
        labelError( ui->labelUserPassword, ui->labelUserPasswordError,
                    tr( "Your passwords do not match!" ) );
        m_readyPassword = false;
    }
    else
    {
        bool ok = true;
        for ( auto pc : m_passwordChecks )
        {
            QString s = pc.filter( pw1 );
            if ( !s.isEmpty() )
            {
                labelError( ui->labelUserPassword, ui->labelUserPasswordError, s );
                ok = false;
                m_readyPassword = false;
                break;
            }
        }

        if ( ok )
        {
            labelOk( ui->labelUserPassword, ui->labelUserPasswordError );
            m_readyPassword = true;
        }
    }

    emit checkReady( isReady() );
}

void
UsersPage::addPasswordCheck( const QString& key, const QVariant& value )
{
    if ( key == "minLength" )
    {
        add_check_minLength( m_passwordChecks, value );
    }
    else if ( key == "maxLength" )
    {
        add_check_maxLength( m_passwordChecks, value );
    }
#ifdef CHECK_PWQUALITY
    else if ( key == "libpwquality" )
    {
        add_check_libpwquality( m_passwordChecks, value );
    }
#endif
    else
        cWarning() << "Unknown password-check key" << key;
}
