/* === This file is part of Calamares - <https://github.com/calamares> ===
 *
 *   Copyright 2014-2015, Teo Mrnjavac <teo@kde.org>
 *   Copyright 2015,      Anke Boersma <demm@kaosx.us>
 *   Copyright 2017-2019, Adriaan de Groot <groot@kde.org>
 *
 *   Calamares is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Calamares is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Calamares. If not, see <http://www.gnu.org/licenses/>.
 */

#include "WelcomePage.h"

#include "ui_WelcomePage.h"
#include "checker/CheckerContainer.h"

#include "Branding.h"
#include "CalamaresVersion.h"
#include "Settings.h"
#include "ViewManager.h"
#include "GlobalStorage.h"
#include "JobQueue.h"

#include "modulesystem/ModuleManager.h"
#include "utils/CalamaresUtilsGui.h"
#include "utils/Logger.h"
#include "utils/Retranslator.h"

#include <QApplication>
#include <QBoxLayout>
#include <QComboBox>
#include <QDesktopServices>
#include <QFocusEvent>
#include <QLabel>
#include <QMessageBox>
#include <QDirIterator>
#include <QFile>
#include <QFileInfo>
#include <QJsonDocument>
#include <QJsonObject>

const QString KeyReadOnlyVerityRoot = "ReadOnlyVerityRoot";
const QString KeyEnableVerityRoot = "Enable";

WelcomePage::WelcomePage( QWidget* parent )
    : QWidget( parent )
    , ui( new Ui::WelcomePage )
    , m_checkingWidget( new CheckerContainer( this ) )
{
    connect( Calamares::ModuleManager::instance(), &Calamares::ModuleManager::requirementsResult, m_checkingWidget, &CheckerContainer::requirementsChecked );
    connect( Calamares::ModuleManager::instance(), &Calamares::ModuleManager::requirementsComplete, m_checkingWidget, &CheckerContainer::requirementsComplete );
    connect( Calamares::ModuleManager::instance(), &Calamares::ModuleManager::requirementsProgress, m_checkingWidget, &CheckerContainer::requirementsProgress );
    ui->setupUi( this );

    ui->verticalLayout->insertSpacing( 1, CalamaresUtils::defaultFontHeight() * 2 );
    initSkus();

    ui->mainText->setAlignment( Qt::AlignCenter );
    ui->mainText->setWordWrap( true );
    ui->mainText->setOpenExternalLinks( true );


    QString message = tr( "<h1>Welcome to the %1 installer.</h1>" );

    ui->mainText->setText( message.arg( *Calamares::Branding::VersionedName ) );
    ui->aboutButton->hide();
    ui->verticalLayout->insertWidget( 3, m_checkingWidget);
}

void
WelcomePage::initSkus()
{
    // Fill the list of skus
    ui->skuWidget->clear();
    ui->skuWidget->setInsertPolicy( QComboBox::InsertAlphabetically );

    const QString skuDir = "/etc/calamares/mariner-skus/";
    QDirIterator it(skuDir, QStringList() << "*.json", QDir::Files);
    while (it.hasNext())
    {
        QString q = it.next();

        // Do not show verity read-only root skus
        if ( !checkVerityReadOnlyRoot( it.filePath() ) )
        {
            ui->skuWidget->addItem( QFileInfo( it.fileName() ).baseName() );
        }
    }
    ui->skuWidget->setCurrentIndex(0);
}


void
WelcomePage::setUpLinks( bool showSupportUrl,
                          bool showKnownIssuesUrl,
                          bool showReleaseNotesUrl )
{
    using namespace Calamares;
    ui->supportButton->hide();
    ui->knownIssuesButton->hide();
    ui->releaseNotesButton->hide();
}


void
WelcomePage::focusInEvent( QFocusEvent* e )
{
    if ( ui->skuWidget )
        ui->skuWidget->setFocus();
    e->accept();
}

bool
WelcomePage::verdict() const
{
    return m_checkingWidget->verdict();
}

QString
WelcomePage::selectedSku() const
{
    return ui->skuWidget->currentText();
}

bool
WelcomePage::checkVerityReadOnlyRoot( const QString &fileName )
{
    bool verityReadOnlyRoot = false;
    QFile file;
    QString val;

    file.setFileName( fileName );
    file.open( QIODevice::ReadOnly );
    val = file.readAll();
    file.close();

    QJsonDocument d = QJsonDocument::fromJson( val.toUtf8() );
    QJsonObject obj = d.object();
    QJsonObject readOnlyObj = obj[KeyReadOnlyVerityRoot].toObject();

    if (readOnlyObj[KeyEnableVerityRoot].toBool() == true)
    {
        verityReadOnlyRoot = true;
    }

    return verityReadOnlyRoot;
}
